.global trampoline, trampoline_end
trampoline:
# We arrive here called by a TB that was trying to 
# emulate a STO instruction whose destination address
# was inside some TB.
#
# We save the registers used by our caller TB before
# calling the selfmodify function. These are scratch
# registers for the C++ compiler, and therefore the
# selfmodify function may clobber them.
	pushq %rdi
	pushq %rsi
	pushq %rcx
	pushq %rdx
# selmodify receives two parameters: a pointer to the
# mbaby env, and the value of the addr field of the STO
# instruction. The env pointer is already in %rdi and
# we received the addr in the low byte of %r8.
	movzbl %r8b, %esi
# a normal call, with its 32b offset, may not be sufficient:
# the trampoline resides in the mmap()ed area, which may be
# very far from the rest of the code.
	movabs $selfmodify, %rax
	callq *%rax
	popq %rdx
	popq %rcx
	popq %rsi
	popq %rdi
# if selfmodify returns 0 (= false), the TB has modified
# some other TB. We can safely continue with its execution.
	cmpq $0, %rax
	je 1f
# otherwise, we need to call emulate and then return to
# the main loop (we cannot return to the TB: it has been
# deleted by selfmodify).
	movabs $_ZN5mbaby7emulateEv, %rax
	call *%rax
# we skip one stack frame to return to our caller's caller
	addq $8, %rsp	# throw away the return address
# now the top of the stack contains the TB frame. We
# continue with a normal epilogue.
	addq $8, %rsp
	leave
1:	ret
trampoline_end:
.global trampoline_opt, trampoline_opt_end
trampoline_opt:
# this is called if the optimized binary translation is in use.
# It works much like the above.
	pushq %rdi
	pushq %rsi
	pushq %rcx
	pushq %rdx
	pushq %r8
	subq $8, %rdi # obtain a pointer to env
	movzbl %r8b, %esi
	movabs $selfmodify, %rax
	callq *%rax
	popq %r8
	popq %rdx
	popq %rcx
	popq %rsi
	popq %rdi
	cmpq $0, %rax
	je 1f
# before calling emulate, we need to update the VM state, which
# is only stored in the host registers during the TB execution.
# We also have to increment CI to let it point to the STO instruction
# (the optimized TB only increments CI at the end). We receive the
# amount to add to CI in the second byte of %r8
	movl %r8d, %eax
	movb %ah, %al
	movb $0, %ah
	addl %eax, %ecx
	subq $8, %rdi		# pointer to env
	movl %ecx, (%rdi)	# update env->CI
	movl %edx, 4(%rdi)	# update env->A
	movabs $_ZN5mbaby7emulateEv, %rax
	call *%rax
# skip one frame
	addq $8, %rsp
# the epilogue is much simpler for the optimized TB
1:	ret
trampoline_opt_end:
